#include "impactapp.h"
#include <windowsx.h>
#include <setupapi.h>
#include <devguid.h>
extern "C"{
#include <hidsdi.h>
#include <hidpi.h>
}
#include <shlwapi.h>	// wnsprintf()

#pragma intrinsic(memcpy,memset,strcpy,strlen)

HINSTANCE hInstance;
#define elemof(x) (sizeof(x)/sizeof((x)[0]))

/******************************
 * Hardwarenahe Kommunikation *
 ******************************/

static SP_CLASSIMAGELIST_DATA ild;

static struct ImpactApparatus{
 HANDLE hDev;
 OVERLAPPED o;		// Zustand eines laufenden Lesezugriffs
 PHIDP_PREPARSED_DATA pd;
 HIDP_CAPS hidcaps;
 bool inreadfile;
 void Enum(HWND hCombo=0,int index=0);
 void open();
 void close();
 int getinput(void*,DWORD);
 int getfeature(void*,DWORD);
 bool setfeature(const void*,DWORD);
 int setoutput(const void*,DWORD);
 BYTE gotinput();	// tut nur die Flags umsortieren!
 BYTE *inputreport;	// gengend groer Transferpuffer
 BYTE poll(Ergebnis&);
 static const float SS;//=0.009922F;	// Schieber-Strich in mm des Messschiebers

// Die folgenden Strukturen sind dieselben wie im Mikrocontroller-Programm
// Alle Reports verwenden eine Report-ID.
// Konstruktoren vermeiden
 struct Report{
  BYTE repid;
  operator BYTE*() {return &repid;}	// Als Ganzes ansprechen
  operator const BYTE*() const {return &repid;}
  struct Stat{
   short min,max,avg,var;	// Minimum, Maximum, Mittelwert, Varianz
  };
 };
 struct RepF:public Report{	// byteLength = 16
  struct F{BYTE
   bk:2,	// Bit 0 = Benutze folgende Bremskraftvorgabe, sonst vom Poti
   tk:2,	// Triggerkopplung DC,AC,?,?
   tq:2,	// Triggerquelle: ADC8 (Kraft), ADC9 (Weg-Poti), ADC6 (Verdrehtaster), ADC1 (frei)
   lk:2;	// Letzter Kanal, bei Summenabtastrate 6/7 MSa/s, fest 0b01 = 2 Kanle
   operator BYTE&(){return*(BYTE*)this;}	// Zugriff in Gnze
  }f;
  short brems;	// Bremskraftvorgabe, 0..4095
  short tv,th;	// Triggerschwellen fr Oszillogramm-Report, 0..4095
  short tp,tg;	// Prtrigger, Gesamtzeit (in Samples, also 7/6 s)
  short knull;	// Kraft-Nullwert (Vorgabe 920, vernderbar)
  short knib;	// (experimentell) zugeordnete A/D-Wandler-Kanle, zurzeit 0x1698
 }repF;

 struct RepI:public Report{	// byteLength = 50
  BYTE flags;	// Bits:
		// 7 = Neues <p>,<k>,<w> nach 100 ms *
		// 6 = Gendertes repF *
		// 5 = Neuer Wert vom Messschieber *
		// 4 = Kraftsto detektiert, neues <ks> *
		// 3:2 = Phase der Triggerung:
		//	0 = Abwarten von <repF.tp> Samples (nur wenn repF.tp>0)
		//	1 = Warten auf erste Triggerbedingung (Unterschreitung <repF.k1>)
		//	2 = Warten auf zweite Triggerbedingung (berschreitung <repF.k2>)
		//	3 = Abwarten von <repF.tg-repF.tp> Samples
		// 1 = frei
		// 0 = Zustand des Verdreh-Tasters
  Stat p,os[4];	// Hand-Poti, Kraft, Weg-Poti, frei, frei
  short ms,ks;	// Messschieber, Kraftsto-Integral
  short km,dw;	// Kraftsto-Maximum, Wegdifferenz
 }repI;

 struct RepT:public Report{	// byteLength = 6868 = 0x1AD4
  struct INFO{BYTE
   nk:3,	// Anzahl Kanle
   tt:1,	// Echter, kein manueller oder (vielleicht) Auto-Trigger
   nr:4;	// Laufende Nummer des Oszillogramms
  }info;
  WORD fill;		// Anzahl Abtastwerte pro Trace
  WORD data[3432];	// Verschachtelte Traces (hnlich WAV-Datei) ~ 4 ms
 }repT;
#if 0 
 struct RepL:public Report{	// byteLength = 4
  BYTE addr;	// Bit 7 = Lesen
  WORD value;	// erstmal ungenutzt
 }repL;

 struct RepReboot:public Report{
  BYTE doit;	// Nur wenn 1
 };
#endif
private:
 int kraftnull() const {return repF.f.tk?repF.tv:repF.knull;}
 double weg_adc2mm(int,bool) const;
public:
 double kraft_adc2N(int,bool) const;	// ADC-Wert in Kraft umrechnen, via linearer Approximation und gleitender Null
}hea;

const float ImpactApparatus::SS=0.009922F;	// Schieber-Strich in mm des Messschiebers

// Fllt Combobox mit Kandidaten (falls mehrere) oder ffnet index
void ImpactApparatus::Enum(HWND hComboEx,int index) {
 COMBOBOXEXITEM cbei;
 TCHAR buf[12];
 int n=index+1;
 if (hComboEx) {
  SendMessage(hComboEx,CBEM_SETIMAGELIST,0,(LPARAM)ild.ImageList);
  cbei.mask=CBEIF_TEXT|CBEIF_IMAGE|CBEIF_SELECTEDIMAGE;
  cbei.pszText=buf;
  cbei.iItem=0;
  SetupDiGetClassImageIndex(&ild,(LPGUID)&GUID_DEVCLASS_HIDCLASS,&cbei.iImage);
  cbei.iSelectedImage=cbei.iImage;
  ComboBox_ResetContent(hComboEx);
  SetWindowText(hComboEx,0);	// falls editierbar: Text lschen
 }

 GUID hidGuid;
 HidD_GetHidGuid(&hidGuid);
 HDEVINFO devs=SetupDiGetClassDevs(&hidGuid,NULL,0,DIGCF_PRESENT|DIGCF_DEVICEINTERFACE);
 if (devs!=INVALID_HANDLE_VALUE) {
  SP_DEVICE_INTERFACE_DATA di;
  for (DWORD i=0; di.cbSize=sizeof di,SetupDiEnumDeviceInterfaces(devs,0,&hidGuid,i,&di); i++) {
   struct{
    SP_DEVICE_INTERFACE_DETAIL_DATA d;
    TCHAR space[MAX_PATH];
   }d;
   d.d.cbSize=sizeof d.d;
   if (SetupDiGetDeviceInterfaceDetail(devs,&di,&d.d,sizeof d,NULL,NULL)) {
    HANDLE h=CreateFile(d.d.DevicePath,GENERIC_READ|GENERIC_WRITE,
      FILE_SHARE_READ|FILE_SHARE_WRITE,NULL,OPEN_EXISTING,FILE_FLAG_OVERLAPPED,NULL);
    if (h!=INVALID_HANDLE_VALUE) {
     PHIDP_PREPARSED_DATA pd;
     if (HidD_GetPreparsedData(h,&pd)) {
      HIDP_CAPS hidcaps;
      if ((int)HidP_GetCaps(pd,&hidcaps)>=0
      && hidcaps.InputReportByteLength==sizeof repI	 //50
      && hidcaps.OutputReportByteLength==2
      && hidcaps.FeatureReportByteLength==sizeof repT) {	// riesig
       if (hComboEx) {
        wsprintf(buf,TEXT("HID%u"),cbei.iItem);
        SendMessage(hComboEx,CBEM_INSERTITEM,0,(LPARAM)&cbei);
        if (cbei.iItem==index) ComboBox_SetCurSel(hComboEx,cbei.iItem);
        cbei.iItem++;
       }else if (!--n) {
        hea.close();
        hDev=h;
        this->pd=pd;
        this->hidcaps=hidcaps;	// Struktur kopieren
        inputreport=new BYTE[hidcaps.InputReportByteLength];
        goto raus;
       }
      }
      HidD_FreePreparsedData(pd);
     }
     CloseHandle(h);
    }
   }
  }
raus:
  SetupDiDestroyDeviceInfoList(devs);
 }
}

int ImpactApparatus::getfeature(void*report,DWORD len) {
 return HidD_GetFeature(hDev,report,len);
}

bool ImpactApparatus::setfeature(const void*report,DWORD len) {
 return !!HidD_SetFeature(hDev,const_cast<void*>(report),len);
}

int ImpactApparatus::setoutput(const void*report,DWORD len) {
 return HidD_SetOutputReport(hDev,const_cast<void*>(report),len);
}

void ImpactApparatus::open() {
 Enum();
 inreadfile=false;
 repF.repid=2;
 getfeature(repF,sizeof repF);
}

void ImpactApparatus::close() {
 if (inputreport) {
  delete[] inputreport; inputreport=0;
 }
 if (pd) {HidD_FreePreparsedData(pd); pd=0;}
 if (hDev) {CloseHandle(hDev); hDev=0;}
}

// Diese Funktion kehrt sofort zurck (blockiert nicht)
// aber guckt nach ob es
BYTE ImpactApparatus::poll(Ergebnis&erg) {
 if (!hDev) return 0;
 BYTE ret=0x80;	// Bit 7 = hDev okay
 DWORD br=(DWORD)-1;
 if (!inreadfile) {
  ZeroMemory(&o,sizeof o);
  if (ReadFile(hDev,inputreport,hidcaps.InputReportByteLength,&br,&o)) {
   ret|=gotinput();	// synchron geklappt
  }else if (GetLastError()==ERROR_IO_PENDING) inreadfile=true;	// nichts weiter tun
 }else if (HasOverlappedIoCompleted(&o)) {
  inreadfile=false;
  ret|=gotinput();	// asynchron geklappt
 }
// Alles andere /stets/ liefern
 if (repI.flags&1) ret|=8;	// Mikrotaster-Stellung im Bit 3
 erg.poti    = repI.p.avg/4095.0F;
 erg.pos     = (float)weg_adc2mm(repI.os[1].avg,true);
 erg.weg     = (float)weg_adc2mm(repI.dw,false);	// ist schon relativ
 erg.stoss   = (float)kraft_adc2N(repI.ks,false)*1E-3F;	// kommt in Nms 'raus, mit 0.001 in Ns umwandeln
 erg.maxkraft= (float)kraft_adc2N(repI.km,false);	// kommt in N raus, also (erstmal) richtig
 return ret;
}	// Deutlich: Mit dem Gestnge ist der Kraftsto viel mehr langgezogen
	// als beim Hmmern auf die blanke Kraftmessdose.
	// Es htte eine geringere Abtastrate (der ATmega32U4) gengt.

double ImpactApparatus::weg_adc2mm(int adcval,bool absolute) const{
// Der linear angenommene Faktor ist 50mm/(3878-39), siehe Wegpoti-HID-Deskriptor
 if (absolute) adcval-=39;	// im Absolutfall theoretisch Umrechnung mit Tabelle
 return adcval*50.0/(3878-39);	// Ergebnis in mm
}

double ImpactApparatus::kraft_adc2N(int adcval, bool absolute) const{	// adcval = A/D-Wert vom Oszillogramm, 0..4095
 if (absolute) adcval-=kraftnull();	// nullzentrierter A/D-Wert
 static const short calib[]={0,778,1464,2004,2283};	// Alle 10 kN eine Sttzstelle, ermittelt 220920
// Intervall finden
 size_t i=0;
 for (; i<elemof(calib)-2; i++) {
  if (adcval < calib[i+1]) break;
 }
 double c0=calib[i], c1=calib[i+1]-c0,
  i_frac = (adcval-c0)/c1;	// normalerweise im Intervall [0..1>
 return (i+i_frac)*1E4F;	// in N (handlicher wren kN gewesen)
}

BYTE ImpactApparatus::gotinput() {
 BYTE ret=0;
 if (inputreport[0]==1) {	// Report-ID
  RepI&cur=*(RepI*)inputreport;
  if ((repI.flags^cur.flags)&1) ret|=2;	// Mikrotaster gendert
  if (cur.flags&0x20) ret|=1;	// Neue Messschieber-Position (ca. 300 ms) - irrelevant, Weg kommt beim Schlag vom Poti
  if (cur.flags&0x40) ret|=32;	// Genderter Feature-Report
  if (cur.flags&0x80) ret|=16;	// Statistikwerte neu (100 ms)
  repI=cur;			// Struktur kopieren
  if (cur.flags&0x10) {		// Kraftsto detektiert
   TCHAR s[80];
   wnsprintf(s,elemof(s),TEXT("Kraftsto: ks = %d, km = %d, dw = %d.\n"),repI.ks,repI.km,repI.dw);
   OutputDebugString(s);
   ret|=4;
  }
 }
 inputreport[0]=0;
 return ret;
}

FUNC(bool) setBrake(float v) {
 if (!hea.getfeature(hea.repF,sizeof hea.repF)) return false;
 if (v!=v) {
  hea.repF.f&=~1;		// Bei NaN auf Potenziometersteuerung zurckstellen
 }else{
  hea.repF.f|=1;
  if (v>1) v=1;
  if (v<0) v=0;
  hea.repF.brems=WORD(v*4095);
 }
 return hea.setfeature(hea.repF,sizeof hea.repF);
}

FUNC(BYTE) Poll(Ergebnis&erg) {
 return hea.poll(erg);
}

FUNC(int) getCurve(float*array,int len,float*dt) {
 hea.getfeature(hea.repT,sizeof hea.repT);
 const int n=hea.repT.fill;
 if (len>n) len=n;
 const int nk=hea.repT.info.nk;
 if (dt) *dt=14*nk/12E6F;	// Abtastperiode in s: 14 Takte bei 12 MHz Takt liefern
 if (array) {
  for (int i=0; i<len; i++) array[i]=(float)hea.kraft_adc2N(hea.repT.data[i*nk],true);
 }
 return len;
}

FUNC(void) Enum(HWND hComboEx,int index) {hea.Enum(hComboEx,index);}

EXTERN_C BOOL CALLBACK _DllMainCRTStartup(HINSTANCE hInst,ULONG reason,void*) {
 switch (reason) {
  case DLL_PROCESS_ATTACH: {
   hInstance=hInst;
   DisableThreadLibraryCalls(hInst);
   ild.cbSize=sizeof(ild);
   SetupDiGetClassImageList(&ild);
   hea.open();
  }break;
  case DLL_PROCESS_DETACH: {
   hea.close();
   SetupDiDestroyClassImageList(&ild);
  }break;
 }
 return TRUE;
}

extern "C" BYTE _fltused;
BYTE _fltused;
